unit filter_f;

interface

type trgbbytes = array[0..2] of byte;
     tdata     = array[0..76799] of trgbbytes;
     tai76803  = array[0..76803] of integer;
     tfilterproc=procedure(var input,output:tdata;i,j:integer);

var filter_job:string[255];
    filter_reco:string[255];
    filter_progress:integer;

procedure filt_mediere(var input,output:tdata;dim,p2:integer);
procedure filt_localtreshold(var input,output:tdata;dim,tresh:integer);
procedure filt_median(var input,output:tdata;dim,p2:integer);
procedure filt_copy(var input,output:tdata;p1,p2:integer);
procedure filt_grayscale(var input,output:tdata;p1,p2:integer);
procedure filt_invert(var input,output:tdata;p1,p2:integer);
procedure filt_afloodbin(var input,output:tdata;treshold,p2:integer);
procedure filt_edge(var input,output:tdata;sensib,medins:integer);
procedure filt_treshold(var input,output:tdata;tr,p2:integer);
procedure filt_recognize(var input,output:tdata;p1,p2:integer);
procedure filt_adaptiveflood(var input,output:tdata;treshold,p2:integer);

const nr_filters=11;
      filternames:array[1..nr_filters] of string[32]=(
        'copy',
        'mediere',
        'median',
        'invert',
        'grayscale',
        'aflood binarize',
        'edge detection',
        'treshold',
        'recognize',
        'adaptive flood',
        'local treshold'
        );
      filterpointer:array[1..nr_filters] of tfilterproc=(
        filt_copy,
        filt_mediere,
        filt_median,
        filt_invert,
        filt_grayscale,
        filt_afloodbin,
        filt_edge,
        filt_treshold,
        filt_recognize,
        filt_adaptiveflood,
        filt_localtreshold
        );
      filterparam:array[1..nr_filters,1..10] of integer=(
      (  0,  0,  0,  0,   0,  0,  0,  0,  0,   0),
      (  2,  2,  2,  5,  32,  0,  0,  0,  0,   0),
      (  2,  2,  2,  5,  32,  0,  0,  0,  0,   0),
      (  0,  0,  0,  0,   0,  0,  0,  0,  0,   0),
      (  0,  0,  0,  0,   0,  0,  0,  0,  0,   0),
      (  0, 20,200,400,1000,  0,300,800,900,1000),
      (  1, 20,200,400,1000,  0,  5,  8, 11,  16),
      (  0, 90,128,156, 255,  0,  0,  0,  0,   0),
      (  0,  0,  0,  0,   0,  0,  0,  0,  0,   0),
      (  0, 20,200,400,1000,  0,  0,  0,  0,   0),
      (  2, 20, 30, 40, 100,  0,205,255,305, 512)
      );

implementation

procedure filt_edge;
const vecin:array[0..35] of integer=(
          -961,-960,-959,
     -642,-641,-640,-639,-638,
-323,-322,-321,-320,-319,-318,-317,
  -3,  -2,  -1,        1,   2,   3,
 317, 318, 319, 320, 321, 322, 323,
      638, 639, 640, 641, 642,
           959, 960, 961);
var x,y,z,i,j,s1,s2:integer;
    cost,cos2:array[-255..255] of integer;
begin
   for x:=-255 to 255 do cost[x]:=round(x*abs(x)/(sqr(x)+sensib)*255);
   for x:=-255 to 255 do cos2[x]:=abs(cost[x]);
   fillchar(output,sizeof(output),0);
   for x:=3 to 236 do for y:=3 to 316 do begin
      z:=x*320+y;
      s1:=0;
      s2:=0;
      for i:=0 to 35 do begin
         j:=z+vecin[i];
         s1:=s1+cost[input[j,0]-input[z,0]];
         s2:=s2+cos2[input[j,0]-input[z,0]];
      end;
      s1:=abs(s1);
      s2:=s2-s1*medins div 16;
      j:=s2 div 36;
      output[z,0]:=j;
      output[z,1]:=j;
      output[z,2]:=j;
   end;
end;

procedure filt_mediere;
var sum:array of integer;
    i,j,k:integer;
begin
   setlength(sum,76800);
   for k:=0 to 2 do begin
      filter_progress:=k*33;
      for i:=0 to 319 do sum[i]:=input[i,k];
      for i:=1 to 239 do sum[i*320]:=sum[i*320-320]+input[i*320,k];

      for i:=1 to 239 do for j:=1 to 319 do begin
         sum[i*320+j] :=  sum[(i-1)*320+j]
                         +sum[i*320+(j-1)]
                         -sum[(i-1)*320+(j-1)]
                         +input[i*320+j,k];
      end;
      for i:=0 to 239-dim do for j:=0 to 319-dim do begin
         output[i*320+j,k]:=round((
              sum[i*320+j]+sum[(i+dim)*320+(j+dim)]-
              sum[(i+dim)*320+j]-sum[i*320+(j+dim)]
              )/(dim*dim));
      end;
      for i:=240-dim to 239 do for j:=0 to 319 do output[i*320+j,k]:=0;
      for i:=0 to 239 do for j:=320-dim to 319 do output[i*320+j,k]:=0;
   end;
   filter_progress:=100;
end;

procedure filt_localtreshold;
var sum:array of integer;
    i,j,q,w,k,dd2,x:integer;
begin
   dd2:=dim div 2;
   setlength(sum,76800);
   tresh:=tresh-256;
   for k:=0 to 2 do begin
      filter_progress:=k*33;
      for i:=0 to 319 do sum[i]:=input[i,k];
      for i:=1 to 239 do sum[i*320]:=sum[i*320-320]+input[i*320,k];

      for i:=1 to 239 do for j:=1 to 319 do begin
         sum[i*320+j] :=  sum[(i-1)*320+j]
                         +sum[i*320+(j-1)]
                         -sum[(i-1)*320+(j-1)]
                         +input[i*320+j,k];
      end;
      for i:=0 to 239 do for j:=0 to 319 do begin
              q:=i-dd2;
              w:=j-dd2;
              if q<0 then q:=0;
              if w<0 then w:=0;
              if q>239-dim then q:=239-dim;
              if w>319-dim then w:=319-dim;
              x:=input[i*320+j,k];
              x:=x-round((sum[q*320+w]+sum[(q+dim)*320+(w+dim)]-
                          sum[(q+dim)*320+w]-sum[q*320+(w+dim)]
                          )/(dim*dim));
              if x<tresh then output[i*320+j,k]:=0
                         else output[i*320+j,k]:=255;
      end;
   end;
   filter_progress:=100;
end;


procedure filt_copy;
begin
   output:=input;
   filter_progress:=100;
end;

procedure filt_grayscale;
var i,j:integer;
begin
   for i:=0 to 76799 do begin
      j:=(input[i,0]+input[i,1]+input[i,2]) div 3;
      output[i,0]:=j;
      output[i,1]:=j;
      output[i,2]:=j;
   end;
   filter_progress:=100;
end;

procedure filt_treshold;
var i,j:integer;
begin
   for i:=0 to 76799 do begin
      j:=255*byte(input[i,0]+input[i,1]+input[i,2]>tr*3);
      output[i,0]:=j;
      output[i,1]:=j;
      output[i,2]:=j;
   end;
   filter_progress:=100;
end;

procedure filt_invert;
var i,k:integer;
begin
   for k:=0 to 2 do for i:=0 to 76799 do output[i,k]:=255-input[i,k];
   filter_progress:=100;
end;

procedure filt_median;
var lst:array of byte;
    mi,mj,i,j,k,x,y,l,r,dpa,dp2:integer;
begin
   dpa:=dim*dim;
   dp2:=dpa div 2;
   setlength(lst,dpa+2);
   for k:=0 to 2 do begin
      for mi:=0 to 240-dim do for mj:=0 to 320-dim do begin
         filter_progress:=(k*240+mi)*100 div 720;
         for x:=0 to dim-1 do for y:=0 to dim-1 do
            lst[x*dim+y]:=input[(mi+x)*320+mj+y,k];
         l:=0;
         r:=dpa-1;
         while l<r do begin
            i:=l; j:=r; x:=lst[(l+r) div 2];
            repeat
               while lst[i]<x do i:=i+1;
               while x<lst[j] do j:=j-1;
               if i<=j then begin
                  y:=lst[i]; lst[i]:=lst[j]; lst[j]:=y;
                  i:=i+1; j:=j-1;
               end;
            until i>j;
            if j>=dp2 then r:=j else if i<=dp2 then l:=i else l:=r;
         end;
         output[mi*320+mj,k]:=lst[dp2];
      end;
      for i:=241-dim to 239 do for j:=0 to 319 do output[i*320+j,k]:=0;
      for i:=0 to 239 do for j:=321-dim to 319 do output[i*320+j,k]:=0;
   end;
   filter_progress:=100;
end;

procedure filt_afloodbin;
const vecin:array[0..7] of integer=(-321,-320,-319,-1,1,319,320,321);
var gnr,qnr,qou,x,y,z,i,j,best,x1,x2,y1,y2:integer;
    isbkgnd:boolean;
    apr:array of integer; {group of pixel}
    que:array of integer; {fill queue}
    gst:array of integer; {group start}
    grc:array of trgbbytes; {neighbour list}
    bkg:array of boolean; {group start}

begin
   setlength(apr,76800);
   fillchar(apr[0],76800*sizeof(integer),0);
   setlength(bkg,76803);
   fillchar(bkg[0],76803*sizeof(boolean),0);
   setlength(grc,76803);
   setlength(gst,76803);
   setlength(que,76803);
   gnr:=2;
   qnr:=0;
   qou:=0;
   gst[1]:=0;
   for z:=0 to 76799 do if apr[z]=0 then begin
      que[qnr]:=z;
      qnr:=qnr+1;
      apr[z]:=gnr;
      grc[z]:=input[z];
      while qnr>qou do begin
         x:=que[qou];
         qou:=qou+1;
         for i:=0 to 7 do begin
            y:=x+vecin[i];
            if (y>=0) and (y<76800) then if apr[y]=0 then
              if sqr(input[y,0]-grc[x,0])+
                 sqr(input[y,1]-grc[x,1])+
                 sqr(input[y,2]-grc[x,2])<
                 treshold then
                begin
                   grc[y,0]:=(input[y,0]+3*grc[x,0]) div 4;
                   grc[y,1]:=(input[y,1]+3*grc[x,1]) div 4;
                   grc[y,2]:=(input[y,2]+3*grc[x,2]) div 4;
                   apr[y]:=gnr;
                   que[qnr]:=y;
                   qnr:=qnr+1;
                end;
         end;
      end;
      inc(gnr);
      gst[gnr]:=qnr;
   end;
   dec(gnr);
   {large group detection}
   for i:=1 to gnr do begin
      z:=gst[i+1]-gst[i];
      {is it too big}
      isbkgnd:=z>p2;
      {did it spread too much}
      x1:=320; x2:=0;
      y1:=240; y2:=0;
      for j:=gst[i] to gst[i+1]-1 do begin
         x:=que[j] mod 320;
         y:=que[j] div 320;
         if x>x2 then x2:=x;
         if x<x1 then x1:=x;
         if y>y2 then y2:=y;
         if y<y1 then y1:=y;
      end;
      isbkgnd:=isbkgnd or (x2-x1>40) or (y2-y1>40);
      {in case we decided this is background}
      if isbkgnd then bkg[i]:=true;
   end;

   for i:=1 to gnr do

      for j:=gst[i] to gst[i+1]-1 do begin
         if bkg[i] then z:=0 else z:=255;
         output[que[j],0]:=z;
         output[que[j],1]:=z;
         output[que[j],2]:=z;
      end;
   filter_progress:=100;
end;

procedure filt_adaptiveflood;
const vecin:array[0..7] of integer=(-321,-320,-319,-1,1,319,320,321);
var gnr,qnr,qou,x,y,z,i,j:integer;
    sumc0,sumc1,sumc2,sumcn:integer;
    apr:array of integer; {group of pixel}
    que:array of integer; {fill queue}
    gst:array of integer; {group start}
    grc:array of trgbbytes; {neighbour list}
    col:array of trgbbytes; {neighbour list}
    bkg:array of boolean; {group start}

begin
   setlength(apr,76800);
   fillchar(apr[0],76800*sizeof(integer),0);
   setlength(bkg,76803);
   fillchar(bkg[0],76803*sizeof(boolean),0);
   fillchar(output,sizeof(tdata),$ffffff);
   setlength(grc,76803);
   setlength(gst,76803);
   setlength(que,76803);
   setlength(col,76803);
   gnr:=1;
   qnr:=0;
   qou:=0;
   gst[1]:=0;
   for z:=0 to 76799 do if apr[z]=0 then begin
      que[qnr]:=z;
      qnr:=qnr+1;
      apr[z]:=gnr;
      grc[z]:=input[z];
      sumc0:=0;
      sumc1:=0;
      sumc2:=0;
      sumcn:=0;
      while qnr>qou do begin
         x:=que[qou];
         sumc0:=sumc0+input[x,0];
         sumc1:=sumc1+input[x,1];
         sumc2:=sumc2+input[x,2];
         qou:=qou+1;
         inc(sumcn);
         for i:=0 to 7 do begin
            y:=x+vecin[i];
            if (y>=0) and (y<76800) then if apr[y]=0 then
              if sqr(input[y,0]-grc[x,0])+
                 sqr(input[y,1]-grc[x,1])+
                 sqr(input[y,2]-grc[x,2])<
                 treshold then
                begin
                   grc[y,0]:=(input[y,0]+3*grc[x,0]) div 4;
                   grc[y,1]:=(input[y,1]+3*grc[x,1]) div 4;
                   grc[y,2]:=(input[y,2]+3*grc[x,2]) div 4;
                   apr[y]:=gnr;
                   que[qnr]:=y;
                   qnr:=qnr+1;
                end;
         end;
      end;
      col[gnr,0]:=sumc0 div sumcn;
      col[gnr,1]:=sumc1 div sumcn;
      col[gnr,2]:=sumc2 div sumcn;
      inc(gnr);
      gst[gnr]:=qnr;
   end;
   dec(gnr);

   for i:=1 to gnr do
      for j:=gst[i] to gst[i+1]-1 do begin
         output[que[j],0]:=col[i,0];
         output[que[j],1]:=col[i,1];
         output[que[j],2]:=col[i,2];
      end;
   filter_progress:=100;
end;

procedure filt_recognize;
const vecin:array[0..7] of integer=(-321,-320,-319,-1,1,319,320,321);
var vcn,min,max,val,state,cnt,gnr,qnr,qou,x,y,z,i,j,x1,y1,xn,q,w:integer;
    apr:array of integer; {group of pixel}
    que:array of integer; {fill queue}
    gst:array of integer; {group start}
    xr1,xd1,yr1,yd1,xs1,ys1:single;
    sig:array[0..100] of integer;
    rez:string[20];
    binf:boolean;
begin
   p1:=2;
   filter_reco:='';
   setlength(apr,76800);
   fillchar(apr[0],76800*sizeof(integer),0);
   setlength(gst,76803);
   setlength(que,76803);
   fillchar(output,sizeof(tdata),0);
   gnr:=2;
   qnr:=0;
   qou:=0;
   gst[1]:=0;
   cnt:=0;
   for z:=0 to 76799 do if (apr[z]=0) and (input[z,0]=255) then begin
      que[qnr]:=z;
      qnr:=qnr+1;
      apr[z]:=gnr;
      x1:=z mod 320;
      y1:=z div 320;
      xn:=1;
      while qnr>qou do begin
         x:=que[qou];
         qou:=qou+1;
         for i:=0 to 7 do begin
            y:=x+vecin[i];
            if (y>=0) and (y<76800) then if apr[y]=0 then
              if input[y,0]=255 then
                begin
                   apr[y]:=gnr;
                   que[qnr]:=y;
                   qnr:=qnr+1;
                   inc(xn);
                   x1:=x1+y mod 320;
                   y1:=y1+y div 320;
                end;
         end;
      end;
      inc(gnr);
      gst[gnr]:=qnr;
      x:=gst[gnr]-gst[gnr-1];
      xr1:=x1/xn;
      yr1:=y1/xn;
      x1:=round(xr1);
      y1:=round(yr1);
      if x>20 then begin
         for j:=gst[gnr-1] to gst[gnr]-1 do begin
            x:=100;
            output[que[j],0]:=x;
            output[que[j],1]:=x;
            output[que[j],2]:=x;
         end;
         for q:=-3 to 3 do for w:=-3 to 3 do if abs(q)=abs(w) then
            if (x1+q>=0) and (x1+q<320) and (y1+w>=0) and (y1+w<240) then
               output[x1+q+(y1+w)*320,2]:=255;
         for q:=0 to 100 do begin
            xd1:=0.3*sin(q/10);
            yd1:=0.3*cos(q/10);
            ys1:=yr1;
            xs1:=xr1;
            w:=0;
            while input[round(xs1)+320*round(ys1),0]=255 do begin
               if w>100 then break;
               inc(w);
               xs1:=xs1+xd1;
               ys1:=ys1+yd1;
            end;
            sig[q]:=w;
         end;
         for q:=0 to 99 do begin
            sig[q]:=(sig[q]+sig[q+1]) div 2;
            if cnt<5 then output[sig[q]*320+q+cnt*50*320,0]:=255;
         end;
         val:=sig[0];
         state:=0;
         binf:=false;
         vcn:=0;
         for q:=1 to 99 do begin
            if state=0 then begin
               if sig[q]>val+p1 then begin
                  state:=1;
                  val:=sig[q];
                  if q>36 then begin
                     if (q<98) or not binf then begin
                        inc(vcn);
                        if q<38 then binf:=true;
                     end;
                  end;
               end;
               if sig[q]<val then val:=sig[q];
            end else begin
               if sig[q]<val-p1 then begin
                  state:=0;
                  val:=sig[q];
               end;
               if sig[q]>val then val:=sig[q];
            end;
            if cnt<5 then output[state*5*320+q+128+cnt*50*320,0]:=255;
         end;
         min:=100; max:=0;
         for q:=0 to 99 do begin
            if sig[q]<min then min:=sig[q];
            if sig[q]>max then max:=sig[q];
         end;
         rez:='';
         if max-min<8 then rez:='cerc' else if max-min>25 then rez:='stea' else begin
            if vcn=3 then rez:='triunghi';
            if vcn=4 then rez:='patrat';
         end;
         filter_reco:=filter_reco+rez+#13#10;
         cnt:=cnt+1;
      end;

   end;
end;

end.
